/*
 FCEU.c

 * MacFCEU 0.6.1
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  fceu.c Copyright (C) 2003 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdarg.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "cart.h"
#include "PPU.h"
#include "NSF.h"
#include "VSUni.h"
#include "FDS.h"
#include "iNES.h"
#include "UNIF.h"
#include "fsound.h"
#include "netplay.h"
#include "general.h"
#include "utility/endian.h"
#include "memory.h"

#include "file.h"
#include "cheat.h"
#include "palette.h"
#include "state.h"
#include "movie.h"
#include "video.h"
#include "input.h"

#include "GameGenie.h"

#ifdef env68k
 #pragma segment shell
#endif

uint8		ConsolePitchFix = 0; /* pitch correction (PAL) */
uint32		EmulationSpeedRatio = 1000000000UL; /* added; 2009.01.29 */

uint64		timestampbase;

FCEUS		FSettings;

FCEUGI		*FCEUGameInfo = 0;

void (*GameInterface)(int h);
void (*GameStateRestore)(int v);

readfunc	ARead[0x10000];
writefunc	BWrite[0x10000];

DECLFW(BNull)
{
(void)A; (void)V;
}

DECLFR(ANull)
{
(void)A;
	return A>>8; //XFrame.DB
}

writefunc FASTAPASS(1) GetWriteHandler(int32 a)
{
	if (GenieRWWrap && a>=0x8000)
		return BWriteG[a-0x8000];
	else
		return BWrite[a];
}

readfunc FASTAPASS(1) GetReadHandler(int32 a)
{
	if (a>=0x8000 && GenieRWWrap)
		return AReadG[a-0x8000];
	else
		return ARead[a];
}

void FASTAPASS(3) SetReadHandler(int32 s,
	int32 e,
	readfunc f)
{
	SetReadHandlerI(1,s,e,f);
}

void FASTAPASS(3) SetWriteHandler(int32 s,
	int32 e,
	writefunc f)
{
	SetWriteHandlerI(1,s,e,f);
}

//NOTE: you have to get the final address right for the custom-incremental ones;
// they set from end
void FASTAPASS(4) SetReadHandlerI(int d,
	int32 start,
	int32 end,
	readfunc f)
{
	int32		a;

	if (f==0)
		f = ANull;

	if (start>(a=end) || (unsigned)start>0xFFFF)
		return;
	if (GenieRWWrap)
		do
			if (a>=0x8000)
				AReadG[a-0x8000] = f;
			else
				ARead[a] = f;
		while((a-=d)>=start);
	else
		do
			ARead[a] = f;
		while((a-=d)>=start);
}

void FASTAPASS(4) SetWriteHandlerI(int d,
	int32 start,
	int32 end,
	writefunc f)
{
	int32		a;

	if (start>(a=end) || (unsigned)start>0xFFFF)
		return;

	if (f==0)
		f = BNull;
	if (GenieRWWrap)
		do
			if (a>=0x8000)
				BWriteG[a-0x8000] = f;
			else
				BWrite[a] = f;
		while((a-=d)>=start);
	else
		do
			BWrite[a] = f;
		while((a-=d)>=start);
}

uint8		GameMemBlock[0x20000]; /*131072*/
uint8		RAM[0x800];
uint8		PAL = 0;

//Dendy (not represented) would be zero
#define CPU_MODEL	PAL ? 2:1

//does a $"00 00 00 00 FF FF FF FF" pattern
void FCEU_MemoryRand(uint8 *d,
	uint32 z)
{
	int	i, x = 0, m = CPU_MODEL ? ~0:0;

	if (d==0)
		return;
	if ((i=z>>2)!=0)
		do
			*d++ = *d++ = *d++ = *d++ = x;
		while(x^=m,--i!=0);
	if (z&2)
		*d++ = *d++ = x;
	if (z&=1)
		*d++ = x;
	if (m)
	{ //init. signature added; 2010.03.17
		i = 8, x = 0277;
		do
			d[7+i] = x;
		while(x+=010,--i!=0);
	}
}

#ifdef env68k
 #pragma segment files

 DECLFW(FCEU_BRAML);
 DECLFW(FCEU_BRAMH);
 DECLFR(FCEU_ARAML);
 DECLFR(FCEU_ARAMH);
#else
static DECLFW(FCEU_BRAML)
{  
	RAM[A] = V;
}

static DECLFW(FCEU_BRAMH)
{
	RAM[A&0x7FF] = V;
}

static DECLFR(FCEU_ARAML)
{
	return RAM[A];
}

static DECLFR(FCEU_ARAMH)
{
	return RAM[A&=0x7FF];
}
#endif

static void CloseGame(void)
{
	if (FCEUGameInfo==0)
		return;
#ifdef NETWORK
	if (FCEUnetplay)
		FCEUD_NetworkClose();
#endif
	if (FCEUGameInfo->name!=0)
		{FCEU_free(FCEUGameInfo->name); FCEUGameInfo->name = 0;}
	if (FCEUGameInfo->type!=GIT_NSF)
		FCEU_FlushGameCheats(0,0);
	if (GameInterface!=0) /* fixes call to NULL crash; 2009.06.27 */
		GameInterface(GI_CLOSE);
	ResetExState(0,0);
	CloseGenie();
	FCEU_free(FCEUGameInfo);
	FCEUGameInfo = 0;

}

void ResetGameLoaded(void)
{
	if (FCEUGameInfo!=0)
		CloseGame();
	GameStateRestore = 0;
	PPU_hook = 0;
	GameHBIRQHook = 0;
	if (GameExpSound.Kill!=0)
		GameExpSound.Kill();
	FCEU_memsetl(&GameExpSound,0,sizeof(GameExpSound));
	MapIRQHook = 0;
	PAL&= 1;
	pale = 0;
}

FCEUGI *FCEUI_LoadGame(const char *name)
{
	FCEUFILE	*fp;
	char		*ipsfn;
	int		result;

	ResetGameLoaded();

	if ((FCEUGameInfo=FCEU_malloc(sizeof(FCEUGI)))!=0)
		FCEU_memsetl(FCEUGameInfo,0,sizeof(FCEUGI));

	FCEUGameInfo->soundchan = 0;
	FCEUGameInfo->soundrate = 0;
	FCEUGameInfo->name = 0;
	FCEUGameInfo->type = GIT_CART;
	FCEUGameInfo->vidsys = GIV_USER;
	FCEUGameInfo->input[0] = FCEUGameInfo->input[1]=-1;
	FCEUGameInfo->inputfc = -1;
	FCEUGameInfo->cspecial = 0;

	FCEU_printf("Loading '%s'...\n\n",
		name);

	GetFileBase(name);

	ipsfn = FCEU_MakeFName(FCEUMKF_IPS,0,0);
	fp = FCEU_fopen(name,ipsfn,"rb",0);
	if (ipsfn)
		FCEU_free(ipsfn);

	if (fp==0)
	{
		FCEU_PrintError("Error opening '%s'!",
			name);
		return 0;
	}

	if ((result=iNESLoad(name,fp))==0
			&& (result=NSFLoad(fp))==0
			&& (result=UNIFLoad(name,fp))==0
			&& (result=FDSLoad(name,fp))==0)
		FCEU_PrintError("The file is corrupt or the format is unsupported.");
	else if (result<0)
	{
		FCEU_PrintError("A major error occurred while loading the file.");
		if (result==-2)
			FCEUD_Message(" Not enough memory.\n");
	}

	FCEU_fclose(fp);
	if (result<=0)
	{
		FCEUD_Message("File closed.\n");
		return 0;
	}

	FCEU_ResetVidSys();

	if (FCEUGameInfo->type!=GIT_NSF)
		if (FSettings.GameGenie)
			OpenGenie();

	PowerNES();
	FCEUSS_CheckStates();
	FCEUMOV_CheckMovies();

	if (FCEUGameInfo->type!=GIT_NSF)
	{
		FCEU_LoadGamePalette();
		FCEU_LoadGameCheats(0);
	}

	FCEU_ResetPalette(-1);
	FCEU_ResetMessages(); // Save state, status messages, etc.

	return FCEUGameInfo;
}


int FCEUI_Initialize(void)
{
	if (!FCEU_InitVirtualVideo())
		return 0;
	FCEU_memsetl(&FSettings,0,sizeof(FSettings));

	FSettings.UsrFirstSLine[0] = 8;
	FSettings.UsrFirstSLine[1] = 0;
	FSettings.UsrLastSLine[0] = 231; /* 239 is recommended... */
	FSettings.UsrLastSLine[1] = 239;
	FSettings.UsrVBLSLines[0] = 20;
	FSettings.UsrVBLSLines[1] = 70;
	FSettings.SoundVolume = 100;
	FSettings.soundrectamp = 256;
	FSettings.soundactivech = 0x1F;
	FCEUPPU_Init();
	X6502_Init();

	return 1;
}

void FCEUI_Kill(void)
{
	FCEU_KillVirtualVideo();
	FCEU_KillGenie();
}

void FCEUI_Emulate(uint8 **pXBuf, int16 **SoundBuf, int32 *SoundBufSize)
{
	FCEU_UpdateInput();
	if (GenieStage!=1)
		FCEU_ApplyPeriodicCheats();
	FCEUPPU_Loop();

	*SoundBufSize = FlushEmulateSound();

	(void)uint64addu32(timestampbase,timestamp);
	timestamp = 0;

	*pXBuf = XBuf;
	GetSoundBuffer(SoundBuf);
}

void FCEUI_CloseGame(void)
{
	CloseGame();
}

void ResetNES(void)
{
	FCEUMOV_AddCommand(FCEUNPCMD_RESET);
	if (FCEUGameInfo==0)
		return;
	GameInterface(GI_RESETM2);
	FCEUSND_Reset();
	FCEUPPU_Reset();
	X6502_Reset();
}

void PowerNES(void) 
{
	FCEUMOV_AddCommand(FCEUNPCMD_POWER);
	if (FCEUGameInfo==0)
		return;

	FCEU_CheatResetRAM();
	FCEU_CheatSetRAM(2,0,RAM);

	GeniePower();

	FCEU_MemoryRand(RAM,0x800);

	SetReadHandler(0,0x7FF,FCEU_ARAML); //slightly faster "low" (no &=0x7FF)
	SetWriteHandler(0,0x7FF,FCEU_BRAML);
	SetReadHandler(0x800,0x1FFF,FCEU_ARAMH);
	SetWriteHandler(0x800,0x1FFF,FCEU_BRAMH);
	SetReadHandler(0x2000,0xFFFF,ANull);
	SetWriteHandler(0x2000,0xFFFF,BNull);

	InitializeInput();
	FCEUSND_Power();
	FCEUPPU_Power();

	/* Have the external game hardware "powered" after the internal NES stuff.  
	   Needed for the NSF code and VS System code.
	*/
        GameInterface(GI_POWER);
	if (FCEUGameInfo->type==GIT_VSUNI)
		FCEU_VSUniPower();

	//zero the timebase and start the CPU
	_int64xset(timestampbase,0);
	X6502_Power();
	FCEU_PowerCheats();
}

void FCEU_ResetVidSys(void)
{
	int		w;

	if (FCEUGameInfo)
	{
		if (FCEUGameInfo->vidsys==GIV_NTSC)
			w = 0;
		else if (FCEUGameInfo->vidsys==GIV_PAL)
			w = 1;
		else /* default (user input) */
			w = FSettings.PAL ? 1:0;

		FCEUPPU_SetVideoSystem(PAL=w);
	}

	SetSoundVariables();
}

void FCEU_printf(const char *format, ...)
{
	char		temp[2048];

	va_list		ap;

	va_start(ap,format);
	vsprintf(temp,format,ap);
	FCEUD_Message(temp);

	va_end(ap);
}

void FCEU_PrintError(const char *format, ...)
{
	char		temp[2048];

	va_list		ap;

	va_start(ap,format);
	vsprintf(temp,format,ap);
	FCEUD_PrintError(temp);

	va_end(ap);
}

void FCEUI_SetRenderedLines(int ntscf,
	int ntscl,
	int palf,
	int pall)
{
	FSettings.UsrFirstSLine[0] = ntscf;
	FSettings.UsrLastSLine[0] = ntscl;
	FSettings.UsrFirstSLine[1] = palf;
	FSettings.UsrLastSLine[1] = pall;

	if (PAL)
	{
		FSettings.FirstSLine = FSettings.UsrFirstSLine[1];
		FSettings.LastSLine = FSettings.UsrLastSLine[1];
	} else
	{
		FSettings.FirstSLine = FSettings.UsrFirstSLine[0];
		FSettings.LastSLine = FSettings.UsrLastSLine[0];
	}
}

void FCEUI_SetVidSystem(int a)
{
	FSettings.PAL = a ? 1:0;

	if (FCEUGameInfo)
		FCEU_ResetVidSys();

	FCEU_ResetPalette(1); /* now calculated regardless; 2009.01.29 */
}

void FCEUI_SetVBLScanlines(int ntscslv,
	int palslv)
{
	FSettings.UsrVBLSLines[0] = ntscslv;
	FSettings.UsrVBLSLines[1] = palslv;

	if (FCEUGameInfo)
		FCEU_ResetVidSys();
}

int FCEUI_GetCurrentVidSystem(int *slstart,
	int *slend,
	int *slvbl)
{
	if (slstart)
		*slstart = FSettings.FirstSLine;
	if (slend)
		*slend = FSettings.LastSLine;
	if (slvbl)
		*slvbl = FSettings.UsrVBLSLines[PAL ? 1:0];

	return PAL;
}

void FCEUI_SetGameGenie(int a)
{
	FSettings.GameGenie = a ? 1:0;
}

void FCEUI_SetSnapName(int a)
{
	FSettings.SnapName = a;
}

#include "utility/ftrig.h"

/* 24 fractional bits */
#define PAL_FPS   838977920 /* 50.006980896 (~50.007) */
#define NTSC_FPS 1008307711 /* 60.099822938 (~60.1) */

/* result is fraction*10000000/64 (or *=156250) */
uint32 FCEUI_GetDesiredFPS(void)
{
	uint32		rate;

	if (PAL)
		rate = PAL_FPS;
	else
		rate = NTSC_FPS;
	/*else if(ConsolePitchFix)
		rate = NTSC_FPS/(M_12R2*M_12R2)+.9; // 53.542855216
	*/

	if (EmulationSpeedRatio<240000000UL)
		EmulationSpeedRatio = 1000000000UL;

	return ufract30mul(rate,EmulationSpeedRatio);
}

/* moved here to reduce compilation time of non-core stuff; 2009.12.29 */

uint8		fceutileview = 0;
extern uint8	PPUrendis, PPUmaxsprites;

void FCEUI_ToggleTileView(void)
{
	fceutileview^= 1;
}

void FCEUI_SetRenderDisable(int sprites,
	int bg)
{
	if (sprites==2)
		PPUrendis^= 1;
	else if (sprites>=0)
		PPUrendis = (PPUrendis&~1)|(sprites ? 1:0);

	if (bg==4)
		PPUrendis^= 6;
	else if (bg==2)
		PPUrendis^= 2;
	else if (bg>=0)
		PPUrendis = (PPUrendis&~2)|(bg ? 2:0);
}

void FCEUI_DisableSpriteLimitation(int a)
{
	PPUmaxsprites = a ? 64:8; /* must be >=8 */
}
